@@ -5,7 +5,7 @@ from .interactive_brokers import InteractiveBrokers
 from .tradier import Tradier
 from .interactive_brokers_rest import InteractiveBrokersREST
 from .example_broker import ExampleBroker
-from .tradeovate import Tradeovate
+from .tradovate import Tradovate
 from .schwab import Schwab
 from .bitunix import Bitunix
 from .projectx import ProjectX
