@@ -1,57 +1,49 @@
-import asyncio
 import datetime
 import logging
-import traceback
-# from asyncio import CancelledError
-from datetime import timezone
-from decimal import Decimal, getcontext
+import os
+from decimal import ROUND_DOWN, Decimal, getcontext
+from typing import Union
 
 from lumibot.data_sources import CcxtData
 from lumibot.entities import Asset, Order, Position
+from termcolor import colored
 
 from .broker import Broker
 
 
-class Ccxt(CcxtData, Broker):
-    """Inherit CcxtData first and all the price market
-    methods than inherits broker
-
+class Ccxt(Broker):
+    """
+    Crypto broker using CCXT.
     """
 
-    def __init__(self, config, max_workers=20, chunk_size=100, connect_stream=False):
-        # Calling init methods
-        CcxtData.__init__(self, config, max_workers=max_workers, chunk_size=chunk_size)
-        Broker.__init__(self, name="ccxt", connect_stream=connect_stream)
+    def __init__(self, config, data_source: CcxtData = None, max_workers=20, chunk_size=100, **kwargs):
+        if data_source is None:
+            data_source = CcxtData(config, max_workers=max_workers, chunk_size=chunk_size)
+        super().__init__(name="ccxt", config=config, data_source=data_source, max_workers=max_workers, **kwargs)
 
-        self.market = "24/7"
+        # Override default market setting for crypto to be 24/7, but still respect config/env if set
+        self.market = (config.get("MARKET") if config else None) or os.environ.get("MARKET") or "24/7"
         self.fetch_open_orders_last_request_time = None
         self.binance_all_orders_rate_limit = 5
+        if not isinstance(self.data_source, CcxtData):
+            raise ValueError(f"Ccxt Broker's Data Source must be of type {CcxtData}")
+        self.api = self.data_source.api
 
     # =========Clock functions=====================
 
     def get_timestamp(self):
-        """Returns the current UNIX timestamp representation from CCXT.
-
-        Parameters
-        ----------
-        None
-        """
-        logging.warning(
-            "The method 'get_time_to_close' is not applicable with Crypto 24/7 markets."
-        )
+        """Returns the current UNIX timestamp representation from CCXT"""
+        logging.warning("The method 'get_time_to_close' is not applicable with Crypto 24/7 markets.")
         return self.api.microseconds() / 1000000
 
     def is_market_open(self):
-        """Not applicable with Crypto 24/7 markets.
+        """The market is always open for Crypto.
 
         Returns
         -------
-        None
+        True
         """
-        logging.warning(
-            "The method 'is_market_open' is not applicable with Crypto 24/7 markets."
-        )
-        return None
+        return True
 
     def get_time_to_open(self):
         """Not applicable with Crypto 24/7 markets.
@@ -60,9 +52,7 @@ class Ccxt(CcxtData, Broker):
         -------
         None
         """
-        logging.warning(
-            "The method 'get_time_to_open' is not applicable with Crypto 24/7 markets."
-        )
+        logging.warning("The method 'get_time_to_open' is not applicable with Crypto 24/7 markets.")
         return None
 
     def get_time_to_close(self):
@@ -72,14 +62,12 @@ class Ccxt(CcxtData, Broker):
         -------
         None
         """
-        logging.warning(
-            "The method 'get_time_to_close' is not applicable with Crypto 24/7 markets."
-        )
+        logging.debug("The method 'get_time_to_close' is not applicable with Crypto 24/7 markets.")
         return None
 
     def is_margin_enabled(self):
         """Check if the broker is using margin trading"""
-        return "margin" in self.api_keys and self.api_keys["margin"] == True
+        return "margin" in self._config and self._config["margin"]
 
     def _fetch_balance(self):
         params = {}
@@ -90,7 +78,7 @@ class Ccxt(CcxtData, Broker):
         return self.api.fetch_balance(params)
 
     # =========Positions functions==================
-    def _get_balances_at_broker(self, quote_asset):
+    def _get_balances_at_broker(self, quote_asset, strategy):
         """Get's the current actual cash, positions value, and total
         liquidation value from ccxt.
 
@@ -110,23 +98,15 @@ class Ccxt(CcxtData, Broker):
         balances = self._fetch_balance()
 
         currency_key = "currency"
-        if self.api.exchangeId in ["coinbasepro", "kucoin", "kraken", "coinbase"]:
+        if self.api.exchangeId in ["coinbasepro", "kucoin", "kraken", "coinbase", "binance", "bitmex"]:
             balances_info = []
-            reserved_keys = ["total", "free", "used", "info", "timestamp", "datetime"]
+            reserved_keys = ["total", "free", "used", "info", "timestamp", "datetime", "debt"]
             for key in balances:
                 if key in reserved_keys:
                     continue
                 bal = balances[key]["total"]
                 if Decimal(bal) != Decimal("0"):
                     balances_info.append({"currency": key, "balance": bal})
-
-        # TODO: Test binance and switch it to the way we do it for coinbasepro and others if possible
-        elif self.api.exchangeId == "binance":
-            balances_info = []
-            for bal in balances["info"]["balances"]:
-                if (Decimal(bal["free"]) + Decimal(bal["locked"])) != Decimal("0"):
-                    balances_info.append(bal)
-            currency_key = "asset"
         else:
             raise NotImplementedError(f"{self.api.exchangeId} not implemented yet.")
 
@@ -152,16 +132,10 @@ class Ccxt(CcxtData, Broker):
             precision_price = self.api.markets[market]["precision"]["price"]
 
             if self.api.exchangeId == "binance":
-                precision_amount = 10 ** -precision_amount
-                precision_price = 10 ** -precision_price
+                precision_amount = 10**-precision_amount
+                precision_price = 10**-precision_price
 
-            # Binance only has `free` and `locked`.
-            if self.api.exchangeId == "binance":
-                total_balance = Decimal(currency_info["free"]) + Decimal(
-                    currency_info["locked"]
-                )
-            else:
-                total_balance = currency_info["balance"]
+            total_balance = currency_info["balance"]
 
             units = Decimal(total_balance)
 
@@ -189,8 +163,9 @@ class Ccxt(CcxtData, Broker):
                 f"the total value of the holdings."
             )
 
-        gross_positions_value = float(positions_value) + float(total_cash_value)
-        net_liquidation_value = float(positions_value) + float(total_cash_value)
+        total_cash_value = float(total_cash_value)
+        gross_positions_value = float(positions_value) + total_cash_value
+        net_liquidation_value = float(positions_value) + total_cash_value
 
         return (total_cash_value, gross_positions_value, net_liquidation_value)
 
@@ -198,18 +173,24 @@ class Ccxt(CcxtData, Broker):
         """parse a broker position representation
         into a position object"""
 
-        if self.api.exchangeId == "binance":
-            symbol = position["asset"]
+        symbol = position["currency"]
+        hold = position["used"]
+        available = position["free"]
+        quantity = Decimal(position["total"])
+
+        # Check if symbol is in the currencies list
+        if symbol not in self.api.currencies:
+            logging.error(
+                f"The symbol {symbol} is not in the currencies list. "
+                f"Please check the symbol and the exchange currencies list."
+            )
+            precision = None
+
+        elif self.api.exchangeId == "binance":
             precision = str(10 ** -self.api.currencies[symbol]["precision"])
-            quantity = Decimal(position["free"]) + Decimal(position["locked"])
-            hold = position["locked"]
-            available = position["free"]
+
         else:
-            symbol = position["currency"]
             precision = str(self.api.currencies[symbol]["precision"])
-            quantity = Decimal(position["total"])
-            hold = position["used"]
-            available = position["free"]
 
         asset = Asset(
             symbol=symbol,
@@ -217,24 +198,27 @@ class Ccxt(CcxtData, Broker):
             precision=precision,
         )
 
-        position_return = Position(
-            strategy, asset, quantity, hold=hold, available=available, orders=orders
-        )
+        position_return = Position(strategy, asset, quantity, hold=hold, available=available, orders=orders)
         return position_return
 
     def _pull_broker_position(self, asset):
         """Given a asset, get the broker representation
         of the corresponding asset"""
-        response = self._pull_broker_positions()["info"][asset.symbol]
-        return response
+
+        position = None
+        if self.api.exchangeId == "binance":
+            for position in self._pull_broker_positions():
+                if position["currency"] == asset.symbol:
+                    return position
+        else:
+            position = self._pull_broker_positions()["info"][asset.symbol]
+        return position
 
     def _pull_broker_positions(self, strategy=None):
         """Get the broker representation of all positions"""
         response = self._fetch_balance()
 
-        if self.api.exchangeId == "binance":
-            return response["info"]["balances"]
-        elif self.api.exchangeId in ["kraken", "kucoin", "coinbasepro", "coinbase"]:
+        if self.api.exchangeId in ["kraken", "kucoin", "coinbasepro", "coinbase", "binance", "bitmex"]:
             balances_info = []
             reserved_keys = [
                 "total",
@@ -243,7 +227,8 @@ class Ccxt(CcxtData, Broker):
                 "info",
                 "timestamp",
                 "datetime",
-                strategy.quote_asset.symbol,
+                "debt",
+                strategy.quote_asset.symbol if strategy else None,
             ]
             for key in response:
                 if key in reserved_keys:
@@ -261,8 +246,49 @@ class Ccxt(CcxtData, Broker):
                                       the exact spelling of the exchangeId."
             )
 
+    def _parse_broker_positions(self, broker_positions, strategy):
+        """parse a list of broker positions into a
+        list of position objects"""
+        result = []
+        for broker_position in broker_positions:
+            new_pos = self._parse_broker_position(broker_position, strategy)
+
+            # Check if the position is not None
+            if new_pos is not None:
+                result.append(new_pos)
+
+        return result
+
+    def _pull_positions(self, strategy):
+        """Get the account positions. return a list of
+        position objects"""
+        response = self._pull_broker_positions(strategy)
+        result = self._parse_broker_positions(response, strategy.name)
+        return result
+
+    def _pull_position(self, strategy, asset):
+        """
+        Pull a single position from the broker that matches the asset and strategy. If no position is found, None is
+        returned.
+
+        Parameters
+        ----------
+        strategy: Strategy
+            The strategy object that placed the order to pull
+        asset: Asset
+            The asset to pull the position for
+
+        Returns
+        -------
+        Position
+            The position object for the asset and strategy if found, otherwise None
+        """
+        response = self._pull_broker_position(asset)
+        result = self._parse_broker_position(response, strategy)
+        return result
+
     # =======Orders and assets functions=========
-    def _parse_broker_order(self, response, strategy_name, strategy_object):
+    def _parse_broker_order(self, response, strategy_name, strategy_object=None):
         """parse a broker order representation
         to an order object"""
         pair = response["symbol"].split("/")
@@ -276,38 +302,42 @@ class Ccxt(CcxtData, Broker):
             response["side"],
             limit_price=response["price"],
             stop_price=response["stopPrice"],
-            time_in_force=response["timeInForce"].lower(),
+            time_in_force=response["timeInForce"].lower() if response["timeInForce"] else None,
             quote=Asset(
                 symbol=pair[1],
                 asset_type="crypto",
             ),
-            type=response["type"] if "type" in response else None,
+            order_type=response["type"] if "type" in response else None,
         )
         order.set_identifier(response["id"])
-        order.update_status(response["status"])
+        order.status = response["status"]
         order.update_raw(response)
         return order
 
-    def _pull_broker_order(self, id):
+    def _pull_broker_order(self, identifier):
         """Get a broker order representation by its id"""
-        open_orders = self._pull_broker_open_orders()
+        open_orders = self._pull_broker_all_orders()
         closed_orders = self._pull_broker_closed_orders()
         all_orders = open_orders + closed_orders
 
-        response = [order for order in all_orders if order["id"] == id]
+        response = [order for order in all_orders if order["id"] == identifier]
 
         return response[0] if len(response) > 0 else None
 
     def _pull_broker_closed_orders(self):
         params = {}
-        if self.is_margin_enabled():
+
+        if self.api.id == "kraken":  # Check if the exchange is Kraken
+            logging.info("Detected Kraken exchange. Not sending params for closed orders.")
+            params = None  # Ensure no parameters are sent
+        elif self.is_margin_enabled():
             params["tradeType"] = "MARGIN_TRADE"
 
         closed_orders = self.api.fetch_closed_orders(params)
 
         return closed_orders
 
-    def _pull_broker_open_orders(self):
+    def _pull_broker_all_orders(self):
         """Get the broker open orders"""
         # For binance api rate limit on calling all orders at once.
         if self.api.exchangeId == "binance":
@@ -315,10 +345,7 @@ class Ccxt(CcxtData, Broker):
             if self.fetch_open_orders_last_request_time is not None:
                 net_rate_limit = (
                     self.binance_all_orders_rate_limit
-                    - (
-                        datetime.datetime.now()
-                        - self.fetch_open_orders_last_request_time
-                    ).seconds
+                    - (datetime.datetime.now() - self.fetch_open_orders_last_request_time).seconds
                 )
                 if net_rate_limit > 0:
                     logging.info(
@@ -329,7 +356,8 @@ class Ccxt(CcxtData, Broker):
             self.fetch_open_orders_last_request_time = datetime.datetime.now()
 
         params = {}
-        if self.is_margin_enabled():
+
+        if self.is_margin_enabled() and self.api.exchangeId != "binance":
             params["tradeType"] = "MARGIN_TRADE"
 
         orders = self.api.fetch_open_orders(params=params)
@@ -340,6 +368,10 @@ class Ccxt(CcxtData, Broker):
         _flatten_order returns a list containing the main order
         and all the derived ones"""
         orders = [order]
+
+        if self.api.exchangeId == "binance":
+            return orders
+
         if "legs" in order._raw and order._raw.legs:
             strategy_name = order.strategy
             for json_sub_order in order._raw.legs:
@@ -351,76 +383,104 @@ class Ccxt(CcxtData, Broker):
     def _submit_order(self, order):
         """Submit an order for an asset"""
 
-        # Orders limited.
-        order_class = ""
-        order_types = ["market", "limit", "stop_limit"]
-        # TODO: Is this actually true?? Try testing this with a bunch of different exchanges.
-        markets_error_message = (
-            f"Only `market`, `limit`, or `stop_limit` orders work "
-            f"with crypto currency markets."
-        )
+        # Check if order has a quantity
+        if not hasattr(order, "quantity") or order.quantity is None:
+            raise ValueError(f"Order {order} does not have a quantity.")
 
-        if order.order_class != order_class:
-            logging.error(
-                f"A compound order of {order.order_class} was entered. "
-                f"{markets_error_message}"
-            )
+        # Check that order quantity is a numeric type
+        if not isinstance(order.quantity, (int, float, Decimal)):
+            raise ValueError(f"Order quantity must be a numeric type, not {type(order.quantity)}")
+
+        # Check if order quantity is greater than 0.
+        if order.quantity <= 0:
+            logging.warning(f"The order {order} was rejected as the order quantity is 0 or less.")
             return
 
-        if order.type not in order_types:
-            logging.error(
-                f"An order type of {order.type} was entered which is not "
-                f"valid. {markets_error_message}"
-            )
+        # Orders limited.
+        order_class = None
+        order_types = ["market", "limit", "stop_limit"]
+        # TODO: Is this actually true?? Try testing this with a bunch of different exchanges.
+        markets_error_message = "Only `market`, `limit`, or `stop_limit` orders work with crypto currency markets."
+
+        if order.order_class != order_class:
+            logging.error(f"A compound order of {order.order_class} was entered. " f"{markets_error_message}")
+            return
+
+        if order.order_type not in order_types:
+            logging.error(f"An order type of {order.order_type} was entered which is not " f"valid. {markets_error_message}")
             return
 
         # Check order within limits.
         market = self.api.markets.get(order.pair, None)
         if market is None:
-            logging.error(
-                f"An order for {order.pair} was submitted. The market for that pair does not exist"
-            )
+            logging.error(f"An order for {order.pair} was submitted. The market for that pair does not exist")
             order.set_error("No market for pair.")
             return order
 
         limits = market["limits"]
         precision = market["precision"]
-        if self.api.exchangeId == "binance" or "kucoin":
-            precision_amount = str(10 ** -precision["amount"])
+        if self.api.exchangeId in ["binance", "kucoin"]:
+            precision_amount = Decimal(str(10 ** -precision["amount"]))
+        elif self.api.exchangeId == "kraken":
+            initial_precision_amount = Decimal(str(precision["amount"]))
+
+            # Remove a few decimal places because Kraken precision amount is wrong and it's causing orders to fail.
+            precision_exp_modifier = 2
+            initial_precision_exp = abs(initial_precision_amount.as_tuple().exponent)
+            new_precision_exp = initial_precision_exp - precision_exp_modifier
+            factor = 10**new_precision_exp
+            precision_amount = Decimal(1) / Decimal(factor)
         else:
-            precision_amount = str(precision["amount"])
+            # Set the precision for the Decimal context
+            getcontext().prec = 8
+            getcontext().rounding = ROUND_DOWN
+            decimal_value = Decimal(precision["amount"])
+            precision_amount = decimal_value.quantize(Decimal("1e-{0}".format(8)), rounding=ROUND_DOWN)
 
         # Convert the amount to Decimal.
         if hasattr(order, "quantity") and getattr(order, "quantity") is not None:
             qty = Decimal(getattr(order, "quantity"))
-            new_qty = qty.quantize(Decimal(precision_amount))
+
+            # Calculate the precision factor as the reciprocal of precision_amount
+            precision_factor = Decimal("1") / precision_amount
+
+            new_qty = (qty * precision_factor).to_integral_value(rounding="ROUND_DOWN") / precision_factor
+
+            if new_qty <= Decimal(0):
+                logging.warning(
+                    f"The order {order} was rejected as the order quantity is 0 or less after rounding down to the exchange minimum precision amount of {precision_amount}."
+                )
+                return
+
             setattr(
                 order,
                 "quantity",
                 new_qty,
             )
-            try:
-                if limits["amount"]["min"] is not None:
-                    assert Decimal(order.quantity) >= Decimal(limits["amount"]["min"])
-            except AssertionError:
-                logging.warning(
-                    f"\nThe order {order} was rejected as the order quantity \n"
-                    f"was less then the minimum allowed for {order.pair}. The minimum order quantity is {limits['amount']['min']} \n"
-                    f"The quantity for this order was {order.quantity} \n"
-                )
-                return
 
-            try:
-                if limits["amount"]["max"] is not None:
-                    assert order.quantity <= limits["amount"]["max"]
-            except AssertionError:
-                logging.warning(
-                    f"\nThe order {order} was rejected as the order quantity \n"
-                    f"was greater then the maximum allowed for {order.pair}. The maximum order "
-                    f"quantity is {limits['amount']['max']} \n"
-                    f"The quantity for this order was {order.quantity} \n"
-                )
-                return
+            # TODO: Remove this if really not needed by several brokers (keeping for now because it's a big change and need to monitor first).
+            # try:
+            #     if limits["amount"]["min"] is not None:
+            #         assert Decimal(order.quantity) >= Decimal(limits["amount"]["min"])
+            # except AssertionError:
+            #     logging.warning(
+            #         f"\nThe order {order} was rejected as the order quantity \n"
+            #         f"was less then the minimum allowed for {order.pair}. The minimum order quantity is {limits['amount']['min']} \n"
+            #         f"The quantity for this order was {order.quantity} \n"
+            #     )
+            #     return
+
+            # try:
+            #     if limits["amount"]["max"] is not None:
+            #         assert order.quantity <= limits["amount"]["max"]
+            # except AssertionError:
+            #     logging.warning(
+            #         f"\nThe order {order} was rejected as the order quantity \n"
+            #         f"was greater then the maximum allowed for {order.pair}. The maximum order "
+            #         f"quantity is {limits['amount']['max']} \n"
+            #         f"The quantity for this order was {order.quantity} \n"
+            #     )
+            #     return
 
         # Convert the price to Decimal.
         for price_type in [
@@ -428,12 +488,12 @@ class Ccxt(CcxtData, Broker):
             "stop_price",
         ]:
             if hasattr(order, price_type) and getattr(order, price_type) is not None:
+                precision_price = Decimal(str(10 ** -precision["price"])
+                                          ) if self.api.exchangeId == "binance" else precision["price"]
                 setattr(
                     order,
                     price_type,
-                    Decimal(getattr(order, price_type)).quantize(
-                        Decimal(str(precision["price"]))
-                    ),
+                    Decimal(getattr(order, price_type)).quantize(Decimal(str(precision_price))),
                 )
             else:
                 continue
@@ -479,9 +539,7 @@ class Ccxt(CcxtData, Broker):
 
             try:
                 if limits["cost"]["max"] is not None:
-                    assert getattr(order, price_type) * order.quantity <= Decimal(
-                        limits["cost"]["max"]
-                    )
+                    assert getattr(order, price_type) * order.quantity <= Decimal(limits["cost"]["max"])
             except AssertionError:
                 logging.warning(
                     f"\nThe order {order} was rejected as the order total cost \n"
@@ -493,10 +551,11 @@ class Ccxt(CcxtData, Broker):
         args = self.create_order_args(order)
 
         params = {}
-        if self.is_margin_enabled():
+
+        if self.is_margin_enabled() and self.api.exchangeId != "binance":
             params["tradeType"] = "MARGIN_TRADE"
 
-        # if self.api.exchangeId == "coinbase" and order.type == "market":
+        # if self.api.exchangeId == "coinbase" and order.order_type == "market":
         #     params["createMarketBuyOrderRequiresPrice"] = False
 
         try:
@@ -506,16 +565,14 @@ class Ccxt(CcxtData, Broker):
 
             response = self.api.create_order(*args, params=params)
             order.set_identifier(response["id"])
-            order.update_status(response["status"])
+            order.status = response["status"]
             order.update_raw(response)
 
         except Exception as e:
             order.set_error(e)
             message = str(e)
-            logging.info(
-                "%r did not go through. The following error occurred: %s"
-                % (order, message)
-            )
+            full_message = f"{order} did not go through. The following error occurred: {message}"
+            logging.info(colored(full_message, "red"))
 
         return order
 
@@ -631,24 +688,22 @@ class Ccxt(CcxtData, Broker):
         broker = self.api.exchangeId
         if broker == "binance":
             params = {}
-            if order.type in ["stop_limit"]:
+            if order.order_type in ["stop_limit"]:
                 params = {
                     "stopPrice": str(order.stop_price),
                 }
                 # Remove items with None values
                 params = {k: v for k, v in params.items() if v}
 
-            order_type_map = dict(
-                market="MARKET", limit="LIMIT", stop_limit="STOP_LOSS_LIMIT"
-            )
+            order_type_map = dict(market="MARKET", limit="LIMIT", stop_limit="STOP_LOSS_LIMIT")
 
             args = [
                 order.pair,
-                order_type_map[order.type],
+                order_type_map[order.order_type],
                 order.side,
                 str(order.quantity),
             ]
-            if order.type in ["limit", "stop_limit"]:
+            if order.order_type in ["limit", "stop_limit"]:
                 args.append(str(order.limit_price))
 
             if len(params) > 0:
@@ -658,7 +713,7 @@ class Ccxt(CcxtData, Broker):
 
         elif broker in ["kraken", "kucoin", "coinbasepro", "coinbase"]:
             params = {}
-            if order.type in ["stop_limit"]:
+            if order.order_type in ["stop_limit"]:
                 params = {
                     "stop": "entry" if order.side == "buy" else "loss",
                     "stop_price": str(order.stop_price),
@@ -676,13 +731,18 @@ class Ccxt(CcxtData, Broker):
 
             args = [
                 order.pair,
-                order_type_map[order.type],
+                order_type_map[order.order_type],
                 order.side,
                 str(order.quantity),  # check this with coinbase.
             ]
-            if order_type_map[order.type] == "limit":
+            if order_type_map[order.order_type] == "limit":
                 args.append(str(order.limit_price))
 
+            # If coinbase, you need to pass the price even with a market order
+            if broker == "coinbase" and order_type_map[order.order_type] == "market":
+                price = self.data_source.get_last_price(order.asset, quote=order.quote)
+                args.append(str(price))
+
             if len(params) > 0:
                 args.append(params)
 
@@ -697,20 +757,31 @@ class Ccxt(CcxtData, Broker):
 
     def cancel_order(self, order):
         """Cancel an order"""
-        response = self.api.cancel_order(order.identifier, order.symbol)
-        if order.identifier == response:
-            order.set_canceled()
+        if self.api.exchangeId == "binance":
+            response = self.api.cancel_order(order.identifier, order.pair)
+            if order.identifier == response["id"]:
+                order.set_canceled()
+        else:
+            response = self.api.cancel_order(order.identifier, order.symbol)
+            if order.identifier == response:
+                order.set_canceled()
 
     def cancel_open_orders(self, strategy):
         """Cancel all open orders at the broker."""
-        for order in self._pull_broker_open_orders():
+        for order in self._pull_broker_all_orders():
             self.api.cancel_order(order["id"], symbol=order["symbol"])
 
+    def _modify_order(self, order: Order, limit_price: Union[float, None] = None,
+                      stop_price: Union[float, None] = None):
+        """
+        Modify an order at the broker. Nothing will be done for orders that are already cancelled or filled. You are
+        only allowed to change the limit price and/or stop price. If you want to change the quantity,
+        you must cancel the order and submit a new one.
+        """
+        raise NotImplementedError("CCXTBroker modify order is not implemented.")
+
     def get_historical_account_value(self):
-        logging.error(
-            "The function get_historical_account_value is not "
-            "implemented yet for CCXT."
-        )
+        logging.error("The function get_historical_account_value is not " "implemented yet for CCXT.")
         return {"hourly": None, "daily": None}
 
     def wait_for_order_registration(self, order):
@@ -752,3 +823,12 @@ class Ccxt(CcxtData, Broker):
             "Waiting for an order execution is not yet implemented in Crypto, "
             "requires streaming. Check the order status at each interval."
         )
+
+    def _get_stream_object(self):
+        pass
+
+    def _register_stream_events(self):
+        pass
+
+    def _run_stream(self):
+        pass
