@@ -11,7 +11,7 @@ from .yahoo_data import YahooData
 from .ccxt_backtesting_data import CcxtBacktestingData
 from .interactive_brokers_rest_data import InteractiveBrokersRESTData
 from .example_broker_data import ExampleBrokerData
-from .tradeovate_data import TradeovateData
+from .tradovate_data import TradovateData
 from .schwab_data import SchwabData
 from .bitunix_data import BitunixData
 from .databento_data import DataBentoData
