@@ -6,3 +6,5 @@ from .dataline import Dataline
 from .order import Order
 from .position import Position
 from .trading_fee import TradingFee
+from .quote import Quote
+from .chains import Chains
