@@ -1,5 +1,7 @@
 import logging
 from datetime import datetime
+from decimal import Decimal
+from typing import Union
 
 import pandas as pd
 
@@ -44,7 +46,7 @@ class Bars:
     get_last_dividend
         Returns the dividend per share value of the last dataframe row
 
-    get_momentum(start=None, end=None)
+    get_momentum(num_periods=1)
         Calculates the global price momentum of the dataframe.
 
     aggregate_bars(frequency)
@@ -101,7 +103,7 @@ class Bars:
         """
         if df.shape[0] == 0:
             logging.warning(f"Unable to get bar data for {asset} {source}")
-        self.df = df
+
         self.source = source.upper()
         self.asset = asset
         if isinstance(asset, tuple):
@@ -111,11 +113,25 @@ class Bars:
         self.quote = quote
         self._raw = raw
 
+        if "dividend" in df.columns:
+            df.loc[:, "price_change"] = df["close"].pct_change(fill_method=None)
+            df.loc[:, "dividend_yield"] = df["dividend"] / df["close"]
+            df.loc[:, "return"] = df["dividend_yield"] + df["price_change"]
+        else:
+            df = df.assign(return_=df["close"].pct_change(fill_method=None))
+            df.rename(columns={"return_": "return"}, inplace=True)
+
+        self.df = df
+
     def __repr__(self):
         return repr(self.df)
 
     def _repr_html_(self):
         return self.df._repr_html_()
+    
+    def __len__(self):
+        """Return the number of bars (rows) in the DataFrame"""
+        return len(self.df)
 
     @classmethod
     def parse_bar_list(cls, bar_list, source, asset):
@@ -159,7 +175,7 @@ class Bars:
 
         return result
 
-    def get_last_price(self):
+    def get_last_price(self) -> Union[float, Decimal, None]:
         """Return the last price of the last bar
 
         Parameters
@@ -168,10 +184,10 @@ class Bars:
 
         Returns
         -------
-        float
+        float, Decimal or None
 
         """
-        return self.df["close"][-1]
+        return self.df["close"].iloc[-1]
 
     def get_last_dividend(self):
         """Return the last dividend of the last bar
@@ -185,9 +201,9 @@ class Bars:
         float
         """
         if "dividend" in self.df.columns:
-            return self.df["dividend"][-1]
+            return self.df["dividend"].iloc[-1]
         else:
-            logging.warning("Unable to find 'dividend' column in bars")
+            logging.debug("Unable to find 'dividend' column in bars")
             return 0
 
     def filter(self, start=None, end=None):
@@ -213,27 +229,17 @@ class Bars:
 
         return df_copy
 
-    def get_momentum(self, start=None, end=None):
-        """Return the momentum (return based on closing prices) of the bars between start and end
-
-        Parameters
-        ----------
-        start : datetime.datetime
-            The start of the range to filter on (inclusive) (default: None)
-
-        end : datetime.datetime
-            The end of the range to filter on (inclusive) (default: None)
-
-        Returns
-        -------
-        float
+    def get_momentum(self, num_periods: int = 1):
         """
-        df_copy = self.filter(start=start, end=end)
-        n_rows = df_copy.shape[0]
-        if n_rows == 0:
-            return 0
-
-        momentum = df_copy["close"].pct_change(n_rows - 1)[-1]
+        Calculate the momentum of the asset over the last num_periods rows. If dividends are provided by the data source,
+        and included in the return calculation, the momentum will be adjusted for dividends.
+        """
+        df_copy = self.df.copy()
+        if "return" in df_copy.columns:
+            period_adj_returns = df_copy['return'].iloc[-num_periods:]
+            momentum = (1 + period_adj_returns).cumprod().iloc[-1] - 1
+        else:
+            momentum = df_copy['close'].pct_change(num_periods).iloc[-1]
         return momentum
 
     def get_total_volume(self, start=None, end=None):
@@ -259,7 +265,7 @@ class Bars:
         volume = df_copy["volume"].sum()
         return volume
 
-    def aggregate_bars(self, frequency):
+    def aggregate_bars(self, frequency, **grouper_kwargs):
         """
         Will convert a set of bars to a different timeframe (eg. 1 min to 15 min)
         frequency (string): The new timeframe that the bars should be in, eg. "15Min", "1H", or "1D"
@@ -280,7 +286,7 @@ class Bars:
         >>> bars = self.get_historical_prices("AAPL", 60, "minute")
         >>> bars_agg = bars.aggregate_bars("15Min")
         """
-        new_df = self.df.groupby(pd.Grouper(freq=frequency)).agg(
+        new_df = self.df.groupby(pd.Grouper(freq=frequency, **grouper_kwargs)).agg(
             {
                 "open": "first",
                 "close": "last",
