@@ -1,12 +1,22 @@
 import datetime
 import logging
+from decimal import Decimal
+from typing import Union
 
 import pandas as pd
 from lumibot import LUMIBOT_DEFAULT_PYTZ as DEFAULT_PYTZ
-from lumibot.tools.helpers import to_datetime_aware
+from lumibot.tools.helpers import parse_timestep_qty_and_unit, to_datetime_aware
 
+from .asset import Asset
 from .dataline import Dataline
 
+# Set the option to raise an error if downcasting is not possible (if available in this pandas version)
+try:
+    pd.set_option('future.no_silent_downcasting', True)
+except (pd._config.config.OptionError, AttributeError):
+    # Option not available in this pandas version, skip it
+    pass
+
 
 class Data:
     """Input and manage Pandas dataframes for backtesting.
@@ -14,12 +24,14 @@ class Data:
     Parameters
     ----------
     asset : Asset Object
-        Asset object to which this data is attached.
+        Asset to which this data is attached.
     df : dataframe
-        Pandas dataframe containing OHLCV etc trade data. Loaded by user
+        Pandas dataframe containing OHLCV etc. trade data. Loaded by user
         from csv.
         Index is date and must be pandas datetime64.
         Columns are strictly ["open", "high", "low", "close", "volume"]
+    quote : Asset Object
+        The quote asset for this data. If not provided, then the quote asset will default to USD.
     date_start : Datetime or None
         Starting date for this data, if not provided then first date in
         the dataframe.
@@ -114,7 +126,6 @@ class Data:
         quote=None,
         timezone=None,
     ):
-
         self.asset = asset
         self.symbol = self.asset.symbol
 
@@ -128,6 +139,12 @@ class Data:
         else:
             self.quote = quote
 
+        # Throw an error if the quote is not an asset object
+        if self.quote is not None and not isinstance(self.quote, Asset):
+            raise ValueError(
+                f"The quote asset for Data must be an Asset object. You provided a {type(self.quote)} object."
+            )
+
         if timestep not in ["minute", "day"]:
             raise ValueError(
                 f"Timestep must be either 'minute' or 'day', the value you enetered ({timestep}) is not currently supported."
@@ -161,9 +178,7 @@ class Data:
         self.df = self.set_date_format(self.df)
         self.df = self.df.sort_index()
 
-        self.trading_hours_start, self.trading_hours_end = self.set_times(
-            trading_hours_start, trading_hours_end
-        )
+        self.trading_hours_start, self.trading_hours_end = self.set_times(trading_hours_start, trading_hours_end)
         self.date_start, self.date_end = self.set_dates(date_start, date_end)
 
         self.df = self.trim_data(
@@ -207,10 +222,7 @@ class Data:
     def columns(self, df):
         # Select columns to use, change to lower case, rename `date` if necessary.
         df.columns = [
-            col.lower()
-            if col.lower() in ["open", "high", "low", "close", "volume"]
-            else col
-            for col in df.columns
+            col.lower() if col.lower() in ["open", "high", "low", "close", "volume"] else col for col in df.columns
         ]
 
         return df
@@ -219,7 +231,7 @@ class Data:
         df.index.name = "datetime"
         df.index = pd.to_datetime(df.index)
         if not df.index.tzinfo:
-            df.index = df.index.tz_localize(DEFAULT_PYTZ)
+            df.index = pd.to_datetime(df.index).tz_localize(DEFAULT_PYTZ)
         elif df.index.tzinfo != DEFAULT_PYTZ:
             df.index = df.index.tz_convert(DEFAULT_PYTZ)
         return df
@@ -228,10 +240,7 @@ class Data:
         # Set the start and end dates of the data.
         for dt in [date_start, date_end]:
             if dt and not isinstance(dt, datetime.datetime):
-                raise TypeError(
-                    f"Start and End dates must be entries as full datetimes. {dt} "
-                    f"was entered"
-                )
+                raise TypeError(f"Start and End dates must be entries as full datetimes. {dt} " f"was entered")
 
         if not date_start:
             date_start = self.df.index.min()
@@ -249,9 +258,7 @@ class Data:
             date_end,
         )
 
-    def trim_data(
-        self, df, date_start, date_end, trading_hours_start, trading_hours_end
-    ):
+    def trim_data(self, df, date_start, date_end, trading_hours_start, trading_hours_end):
         # Trim the dataframe to match the desired backtesting dates.
 
         df = df.loc[(df.index >= date_start) & (df.index <= date_end), :]
@@ -260,7 +267,7 @@ class Data:
         if df.empty:
             raise ValueError(
                 f"When attempting to load a dataframe for {self.asset}, "
-                f"and empty dataframe was returned. This is likely due "
+                f"an empty dataframe was returned. This is likely due "
                 f"to your backtesting start and end dates not being "
                 f"within the start and end dates of the data provided. "
                 f"\nPlease check that your at least one of your start "
@@ -269,25 +276,52 @@ class Data:
             )
         return df
 
+    # ./lumibot/build/__editable__.lumibot-3.1.14-py3-none-any/lumibot/entities/data.py:280:
+    # FutureWarning: Downcasting object dtype arrays on .fillna, .ffill, .bfill is deprecated and will change in a future version.
+    # Call result.infer_objects(copy=False) instead.
+    # To opt-in to the future behavior, set `pd.set_option('future.no_silent_downcasting', True)`
+
     def repair_times_and_fill(self, idx):
         # Trim the global index so that it is within the local data.
         idx = idx[(idx >= self.datetime_start) & (idx <= self.datetime_end)]
+        
+        # Ensure that the DataFrame's index is unique by dropping duplicate timestamps.
+        self.df = self.df[~self.df.index.duplicated(keep='first')]
+        
+        # Reindex the DataFrame with the new index and forward-fill missing values.
+        df = self.df.reindex(idx, method="ffill")
+        
+        # Check if we have a volume column, if not then add it and fill with 0 or NaN.
+        if "volume" in df.columns:
+            df.loc[df["volume"].isna(), "volume"] = 0
+        else:
+            df["volume"] = None
 
-        # After all time series merged, adjust the local dataframe to reindex and fill nan's.
-        df = self.df.reindex(idx)
-        df.loc[df["volume"].isna(), "volume"] = 0
+        # Forward fill all columns except for open, high, and low.
         df.loc[:, ~df.columns.isin(["open", "high", "low"])] = df.loc[
             :, ~df.columns.isin(["open", "high", "low"])
         ].ffill()
+
+        # If any of close, open, high, low columns are missing, add them with NaN.
+        for col in ["close", "open", "high", "low"]:
+            if col not in df.columns:
+                df[col] = None
+
+        # If there are any NaNs in open, high, or low, fill them with the close value.
         for col in ["open", "high", "low"]:
-            df.loc[df[col].isna(), col] = df.loc[df[col].isna(), "close"]
+            try:
+                df.loc[df[col].isna(), col] = df.loc[df[col].isna(), "close"]
+            except Exception as e:
+                logging.error(f"Error filling {col} column: {e}")
 
         self.df = df
 
+        # Set up iter_index and iter_index_dict for later use.
         iter_index = pd.Series(df.index)
         self.iter_index = pd.Series(iter_index.index, index=iter_index)
         self.iter_index_dict = self.iter_index.to_dict()
 
+        # Populate the datalines dictionary (assuming to_datalines is defined elsewhere).
         self.datalines = dict()
         self.to_datalines()
 
@@ -324,6 +358,10 @@ class Data:
         # known data (this speeds up the process)
         i = None
 
+        # Check if we have the iter_index_dict, if not then repair the times and fill (which will create the iter_index_dict)
+        if getattr(self, "iter_index_dict", None) is None:
+            self.repair_times_and_fill(self.df.index)
+
         # Search for dt in self.iter_index_dict
         if dt in self.iter_index_dict:
             i = self.iter_index_dict[dt]
@@ -338,31 +376,35 @@ class Data:
         # will return data. Runs function if data, returns None if no data.
         def checker(self, *args, **kwargs):
             if type(kwargs.get("length", 1)) not in [int, float]:
-                raise TypeError(
-                    f"Length must be an integer. {type(kwargs.get('length', 1))} was provided."
-                )
+                raise TypeError(f"Length must be an integer. {type(kwargs.get('length', 1))} was provided.")
 
             dt = args[0]
+
             # Check if the iter date is outside of this data's date range.
-            if dt < self.datetime_start or dt > self.datetime_end:
+            if dt < self.datetime_start:
                 raise ValueError(
-                    f"The date you are looking for ({dt}) for ({self.asset}) is outside of the data's date range ({self.datetime_start} to {self.datetime_end})."
+                    f"The date you are looking for ({dt}) for ({self.asset}) is outside of the data's date range ({self.datetime_start} to {self.datetime_end}). This could be because the data for this asset does not exist for the date you are looking for, or something else."
                 )
 
             # Search for dt in self.iter_index_dict
+            if getattr(self, "iter_index_dict", None) is None:
+                self.repair_times_and_fill(self.df.index)
+
             if dt in self.iter_index_dict:
                 i = self.iter_index_dict[dt]
             else:
                 # If not found, get the last known data
                 i = self.iter_index.asof(dt)
 
-            data_index = i + 1 - kwargs.get("length", 1) - kwargs.get("timeshift", 0)
+            length = kwargs.get("length", 1)
+            timeshift = kwargs.get("timeshift", 0)
+            data_index = i + 1 - length - timeshift
             is_data = data_index >= 0
             if not is_data:
-                raise ValueError(
+                # Log a warning
+                logging.warning(
                     f"The date you are looking for ({dt}) is outside of the data's date range ({self.datetime_start} to {self.datetime_end}) after accounting for a length of {kwargs.get('length', 1)} and a timeshift of {kwargs.get('timeshift', 0)}. Keep in mind that the length you are requesting must also be available in your data, in this case we are {data_index} rows away from the data you need."
                 )
-                return None
 
             res = func(self, *args, **kwargs)
             # print(f"Results last price: {res}")
@@ -371,29 +413,7 @@ class Data:
         return checker
 
     @check_data
-    def is_tradable(self, dt, length=1, timestep="minute", timeshift=0):
-        """Returns True if the data is tradeable.
-
-        Parameters
-        ----------
-        dt : datetime.datetime
-            The datetime to check for tradeability.
-        length : int
-            The number of periods to check for tradeability.
-        timestep : str
-            The frequency of the data to check for tradeability.
-        timeshift : int
-            The number of periods to shift the data.
-
-        Returns
-        -------
-        bool
-        """
-        # Return true if data is available for trading. None if not.
-        return True
-
-    @check_data
-    def get_last_price(self, dt, length=1, timeshift=0):
+    def get_last_price(self, dt, length=1, timeshift=0) -> Union[float, Decimal, None]:
         """Returns the last known price of the data.
 
         Parameters
@@ -409,9 +429,94 @@ class Data:
 
         Returns
         -------
-        float
+        float or Decimal or None
         """
-        return self.datalines["open"].dataline[self.get_iter_count(dt)]
+        iter_count = self.get_iter_count(dt)
+        open_price = self.datalines["open"].dataline[iter_count]
+        close_price = self.datalines["close"].dataline[iter_count]
+        price = close_price if dt > self.datalines["datetime"].dataline[iter_count] else open_price
+        return price
+
+    @check_data
+    def get_quote(self, dt, length=1, timeshift=0):
+        """Returns the last known price of the data.
+
+        Parameters
+        ----------
+        dt : datetime.datetime
+            The datetime to get the last price.
+        length : int
+            The number of periods to get the last price.
+        timestep : str
+            The frequency of the data to get the last price.
+        timeshift : int
+            The number of periods to shift the data.
+
+        Returns
+        -------
+        dict
+        """
+        # Check if this data object at least has open, high, low, close, and volume columns
+        if not all(
+            [
+                "open" in self.datalines,
+                "high" in self.datalines,
+                "low" in self.datalines,
+                "close" in self.datalines,
+                "volume" in self.datalines,
+            ]
+        ):
+            raise ValueError(
+                f"The data object for {self.asset} does not have the necessary columns to get the quote. Please make sure that the data object has at least the following columns: open, high, low, close, and volume. This could be an issue with the data source or the data itself, consider changing the data source you are using or check that the data you are looking for exists in the data source."
+            )
+        
+        # Check if this data object has bid and ask
+        if not all(
+            [
+                "bid" in self.datalines,
+                "ask" in self.datalines,
+                "bid_size" in self.datalines,
+                "bid_condition" in self.datalines,
+                "bid_exchange" in self.datalines,
+                "ask_size" in self.datalines,
+                "ask_condition" in self.datalines,
+                "ask_exchange" in self.datalines,
+            ]
+        ):
+            raise ValueError(
+                f"The data object for {self.asset} does not have the necessary columns to get the quote. Please make sure that the data object has at least the following columns: bid and ask. This could be an issue with the data source or the data itself, consider changing the data source you are using or check that the data you are looking for exists in the data source. For example, Polygon does not provide bid and ask data."
+            )
+
+        iter_count = self.get_iter_count(dt)
+        open = round(self.datalines["open"].dataline[iter_count], 2)
+        high = round(self.datalines["high"].dataline[iter_count], 2)
+        low = round(self.datalines["low"].dataline[iter_count], 2)
+        close = round(self.datalines["close"].dataline[iter_count], 2)
+        bid = round(self.datalines["bid"].dataline[iter_count], 2)
+        ask = round(self.datalines["ask"].dataline[iter_count], 2)
+        volume = round(self.datalines["volume"].dataline[iter_count], 0)
+        bid_size = round(self.datalines["bid_size"].dataline[iter_count], 0)
+        bid_condition = round(self.datalines["bid_condition"].dataline[iter_count], 0)
+        bid_exchange = round(self.datalines["bid_exchange"].dataline[iter_count], 0)
+        ask_size = round(self.datalines["ask_size"].dataline[iter_count], 0)
+        ask_condition = round(self.datalines["ask_condition"].dataline[iter_count], 0)
+        ask_exchange = round(self.datalines["ask_exchange"].dataline[iter_count], 0)
+
+        return {
+            "open": open,
+            "high": high,
+            "low": low,
+            "close": close,
+            "volume": volume,
+            "bid": bid,
+            "ask": ask,
+            "bid_size": bid_size,
+            "bid_condition": bid_condition,
+            "bid_exchange": bid_exchange,
+            "ask_size": ask_size,
+            "ask_condition": ask_condition,
+            "ask_exchange": ask_exchange
+        }
 
     @check_data
     def _get_bars_dict(self, dt, length=1, timestep=None, timeshift=0):
@@ -437,17 +542,55 @@ class Data:
         # Get bars.
         end_row = self.get_iter_count(dt) - timeshift
         start_row = end_row - length
+
         if start_row < 0:
             start_row = 0
 
+        # Cast both start_row and end_row to int
+        start_row = int(start_row)
+        end_row = int(end_row)
+
         dict = {}
         for dl_name, dl in self.datalines.items():
             dict[dl_name] = dl.dataline[start_row:end_row]
 
         return dict
 
-    def get_bars(self, dt, length=1, timestep=MIN_TIMESTEP, timeshift=0, exchange=None):
-        """Returns a dictionary of the data.
+    def _get_bars_between_dates_dict(self, timestep=None, start_date=None, end_date=None):
+        """Returns a dictionary of all the data available between the start and end dates.
+
+        Parameters
+        ----------
+        timestep : str
+            The frequency of the data to get the data.
+        start_date : datetime.datetime
+            The start date to get the data for.
+        end_date : datetime.datetime
+            The end date to get the data for.
+
+        Returns
+        -------
+        dict
+        """
+
+        end_row = self.get_iter_count(end_date)
+        start_row = self.get_iter_count(start_date)
+
+        if start_row < 0:
+            start_row = 0
+
+        # Cast both start_row and end_row to int
+        start_row = int(start_row)
+        end_row = int(end_row)
+
+        dict = {}
+        for dl_name, dl in self.datalines.items():
+            dict[dl_name] = dl.dataline[start_row:end_row]
+
+        return dict
+
+    def get_bars(self, dt, length=1, timestep=MIN_TIMESTEP, timeshift=0):
+        """Returns a dataframe of the data.
 
         Parameters
         ----------
@@ -465,30 +608,93 @@ class Data:
         pandas.DataFrame
 
         """
+        # Parse the timestep
+        quantity, timestep = parse_timestep_qty_and_unit(timestep)
+        num_periods = length
+
         if timestep == "minute" and self.timestep == "day":
-            raise ValueError(
-                "You are requesting minute data from a daily data source. This is not supported."
-            )
+            raise ValueError("You are requesting minute data from a daily data source. This is not supported.")
 
         if timestep != "minute" and timestep != "day":
-            raise ValueError(
-                f"Only minute and day are supported for timestep. You provided: {timestep}"
-            )
+            raise ValueError(f"Only minute and day are supported for timestep. You provided: {timestep}")
 
+        agg_column_map = {
+            "open": "first",
+            "high": "max",
+            "low": "min",
+            "close": "last",
+            "volume": "sum",
+        }
         if timestep == "day" and self.timestep == "minute":
             # If the data is minute data and we are requesting daily data then multiply the length by 1440
             length = length * 1440
+            unit = "D"
+            data = self._get_bars_dict(dt, length=length, timestep="minute", timeshift=timeshift)
 
-            dict = self._get_bars_dict(
-                dt, length=length, timestep="minute", timeshift=timeshift
-            )
+        elif timestep == 'day' and self.timestep == 'day':
+            unit = "D"
+            data = self._get_bars_dict(dt, length=length, timestep=timestep, timeshift=timeshift)
+
+        else:
+            unit = "min"  # Guaranteed to be minute timestep at this point
+            length = length * quantity
+            data = self._get_bars_dict(dt, length=length, timestep=timestep, timeshift=timeshift)
+
+        if data is None:
+            return None
+
+        df = pd.DataFrame(data).assign(datetime=lambda df: pd.to_datetime(df['datetime'])).set_index('datetime')
+        if "dividend" in df.columns:
+            agg_column_map["dividend"] = "sum"
+        df_result = df.resample(f"{quantity}{unit}").agg(agg_column_map)
+
+        # Drop any rows that have NaN values (this can happen if the data is not complete, eg. weekends)
+        df_result = df_result.dropna()
+
+        # Remove partial day data from the current day, which can happen if the data is in minute timestep.
+        if timestep == "day" and self.timestep == "minute":
+            df_result = df_result[df_result.index < dt.replace(hour=0, minute=0, second=0, microsecond=0)]
+
+        # The original df_result may include more rows when timestep is day and self.timestep is minute.
+        # In this case, we only want to return the last n rows.
+        df_result = df_result.tail(n=int(num_periods))
+
+        return df_result
+
+    def get_bars_between_dates(self, timestep=MIN_TIMESTEP, exchange=None, start_date=None, end_date=None):
+        """Returns a dataframe of all the data available between the start and end dates.
+
+        Parameters
+        ----------
+        timestep : str
+            The frequency of the data to get the data. Only minute and day are supported.
+        exchange : str
+            The exchange to get the data for.
+        start_date : datetime.datetime
+            The start date to get the data for.
+        end_date : datetime.datetime
+            The end date to get the data for.
+
+        Returns
+        -------
+        pandas.DataFrame
+        """
+
+        if timestep == "minute" and self.timestep == "day":
+            raise ValueError("You are requesting minute data from a daily data source. This is not supported.")
+
+        if timestep != "minute" and timestep != "day":
+            raise ValueError(f"Only minute and day are supported for timestep. You provided: {timestep}")
+
+        if timestep == "day" and self.timestep == "minute":
+            dict = self._get_bars_between_dates_dict(timestep=timestep, start_date=start_date, end_date=end_date)
 
             if dict is None:
                 return None
 
             df = pd.DataFrame(dict).set_index("datetime")
 
-            df_dict = df.resample("D").agg(
+            df_result = df.resample("D").agg(
                 {
                     "open": "first",
                     "high": "max",
@@ -498,13 +704,10 @@ class Data:
                 }
             )
 
-            dict = df_dict.to_dict(orient="list")
+            return df_result
 
-            return df
         else:
-            dict = self._get_bars_dict(
-                dt, length=length, timestep=timestep, timeshift=timeshift
-            )
+            dict = self._get_bars_between_dates_dict(timestep=timestep, start_date=start_date, end_date=end_date)
 
             if dict is None:
                 return None
