@@ -20,13 +20,29 @@ class Position:
         The quantity of the asset owned.
     orders : list of Order
         The orders that have been executed for this position.
+    hold : float
+        The assets that are not free in the portfolio. (Crypto: only)
+    available : float
+        The assets that are free in the portfolio. (Crypto: only)
+    avg_fill_price : float
+        The average fill price of the position.
     """
 
-    def __init__(self, strategy, asset, quantity, orders=None, hold=0, available=0):
+    def __init__(
+            self, 
+            strategy, 
+            asset, 
+            quantity, 
+            orders=None, 
+            hold=0, 
+            available=0, 
+            avg_fill_price=None
+        ):
         self.strategy = strategy
         self.asset = asset
         self.symbol = self.asset.symbol
         self.orders = None
+        self.avg_fill_price = avg_fill_price
 
         # Quantity is the total number of shares/units owned in the position.
         # setting the quantity
@@ -57,12 +73,17 @@ class Position:
             self.orders = orders
 
     def __repr__(self):
-        repr = "%f shares of %s" % (self.quantity, self.asset)
-        return repr
+        return f"{self.strategy} Position: {self.quantity} shares of {self.asset} ({len(self.orders)} orders)"
 
     @property
     def quantity(self):
-        return self._quantity
+        result = float(self._quantity)
+
+        # If result is less than 0.000001, return 0.0 to avoid rounding errors.
+        if abs(result) < 0.000001:
+            return 0.0
+
+        return result
 
     @quantity.setter
     def quantity(self, value):
@@ -134,6 +155,12 @@ class Position:
             An order that can be used to sell this position.
 
         """
+        # Prevent use for crypto futures
+        if getattr(self.asset, "asset_type", None) == "crypto_future":
+            import logging
+            logger = logging.getLogger(__name__)
+            logger.warning("get_selling_order is not supported for crypto futures. Use the broker's close_position method instead.")
+            return None
         order = None
         if self.quantity < 0:
             order = entities.Order(
@@ -145,8 +172,33 @@ class Position:
             )
         return order
 
-    def add_order(self, order, quantity):
-        increment = quantity if order.side == order.BUY else -quantity
-        self.quantity += increment
+    def add_order(self, order: entities.Order, quantity: Decimal = Decimal(0)):
+        increment = quantity if order.side == "buy" else -quantity
+        self._quantity += Decimal(increment)
         if order not in self.orders:
             self.orders.append(order)
+
+    # ========= Serialization methods ===========
+    def to_dict(self):
+        return {
+            "strategy": self.strategy,
+            "asset": self.asset.to_dict(),
+            "quantity": float(self.quantity),
+            "orders": [order.to_dict() for order in self.orders],
+            "hold": self.hold,
+            "available": float(self.available) if self.available else None,
+            "avg_fill_price": float(self.avg_fill_price) if self.avg_fill_price else None,
+        }
+    
+    @classmethod
+    def from_dict(cls, data):
+        asset = entities.Asset.from_dict(data["asset"])
+        return cls(
+            strategy=data["strategy"],
+            asset=asset,
+            quantity=Decimal(data["quantity"]),
+            orders=[entities.Order.from_dict(order) for order in data["orders"]],
+            hold=Decimal(data["hold"]),
+            available=Decimal(data["available"]),
+            avg_fill_price=Decimal(data["avg_fill_price"]),
+        )
\ No newline at end of file
